// Copyright (C) 2009 Playedit.

// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include "config.h"

#include "mtp_device.h"
#include "playlist.h"
#include "music.h"

#include <glibmm/i18n.h>
#include <gtkmm/main.h>
#include <iostream>

PlaylistView::PlaylistView(Playedit::App& app_window)
	: DNDTreeViewBase(_("Playlists")),
	m_UIManager(app_window.m_UIManager),
	m_Device(app_window.m_Device),
	m_bScrollOutOfPlaylist(true)
{
	m_Column.pack_start(m_PlaylistName);
	m_Column.set_cell_data_func(m_PlaylistName,
			sigc::mem_fun(*this, &PlaylistView::PlaylistNameDataFunc));

	// Set up row selections
	get_selection()->set_select_function(sigc::mem_fun(*this, &PlaylistView::SelectionFunc));
	get_selection()->set_mode(Gtk::SELECTION_MULTIPLE);

	// Rename playlist callbacks
	m_PlaylistName.signal_edited().
		connect(sigc::mem_fun(*this, &PlaylistView::RenamePlaylist));
	m_PlaylistName.signal_editing_canceled().
		connect(sigc::mem_fun(*this, &PlaylistView::CancelRenamePlaylist));

	// Setup as a drag destination
	std::vector<Gtk::TargetEntry> target_entries;
	target_entries.push_back(Gtk::TargetEntry("GTK_TREE_MODEL_ROWS", Gtk::TARGET_SAME_APP, 0));
	drag_dest_set(target_entries, Gtk::DEST_DEFAULT_ALL, Gdk::ACTION_COPY | Gdk::ACTION_MOVE);

	// Setup row changed signal to set visibility of the appropriate actions
	// and do it for the first time.
	get_selection()->signal_changed().connect(sigc::mem_fun(*this, &PlaylistView::SelectionChanged));
	SelectionChanged();
}

// Set the text that is shown in the playlist treeview.
void PlaylistView::PlaylistNameDataFunc(Gtk::CellRenderer *renderer,
	const Gtk::TreeIter& iter)
{
	((Gtk::CellRendererText*)renderer)->property_text() =
		iter->get_value(m_StoreColumns.m_Data)->Name();
}

// Add the playlists and tracks to the playlist store so they will be
// displayed.
void PlaylistView::PopulateStore(
		const std::vector< Glib::RefPtr<playlist_t> >& playlist_list,
		MusicView* music_view)
{
	Glib::RefPtr<Gtk::TreeModel> music_store = music_view->get_model();

	for (std::vector< Glib::RefPtr<playlist_t> >::const_iterator playlist = playlist_list.begin();
			playlist != playlist_list.end(); ++playlist)
	{
		Gtk::TreeModel::Row playlist_row = *(m_Store->append());
		playlist_row[m_StoreColumns.m_Data] = *playlist;

		std::vector<uint32_t> track_ids = (*playlist)->Tracks();
		for (std::vector<uint32_t>::iterator track_id = track_ids.begin();
				track_id != track_ids.end(); ++track_id)
		{
			bool found = false;
			for (Gtk::TreeIter album = music_store->children().begin();
					!found && album != music_store->children().end(); ++album)
			{
				for (Gtk::TreeIter iter = album->children().begin();
						!found && iter != album->children().end(); ++iter)
				{
					Glib::RefPtr<track_t> track = Glib::RefPtr<track_t>::cast_static(
							iter->get_value(m_StoreColumns.m_Data));
					if (track->Id() == *track_id)
					{
						Gtk::TreeModel::Row track_row =
							*(m_Store->append(playlist_row.children()));
						track_row[m_StoreColumns.m_Data] = track;
						found = true;
					}
				}
			}
		}
	}
}

// Remove all references to a track from the playlist view.  Called when an
// track is deleted from the device.
void PlaylistView::ClearTrack(const uint32_t track_id)
{
	for (Gtk::TreeIter playlist_iter = m_Store->children().begin();
			playlist_iter != m_Store->children().end(); ++playlist_iter)
	{
		Glib::RefPtr<playlist_t> playlist =
			Glib::RefPtr<playlist_t>::cast_static(playlist_iter->get_value(m_StoreColumns.m_Data));
		std::vector<uint32_t> playlist_tracks = playlist->Tracks();

		for (std::vector<uint32_t>::iterator track_id_iter = playlist_tracks.begin();
				track_id_iter != playlist_tracks.end(); ++track_id_iter)
			if (*track_id_iter == track_id)
				track_id_iter = --playlist_tracks.erase(track_id_iter);
		playlist->SetTracks(playlist_tracks);

		std::vector<Gtk::TreeRowReference> track_refs = ConvertItersToRowRefs(playlist_iter->children());
		for (std::vector<Gtk::TreeRowReference>::iterator track_ref = track_refs.begin();
				track_ref != track_refs.end(); ++track_ref)
		{
			Gtk::TreeIter track_iter = m_Store->get_iter(track_ref->get_path());
			Glib::RefPtr<track_t> track =
				Glib::RefPtr<track_t>::cast_static(track_iter->get_value(m_StoreColumns.m_Data));
			if (track_id == track->Id())
				m_Store->erase(track_iter);
		}
	}
}

bool PlaylistView::SelectionFunc(const Glib::RefPtr<Gtk::TreeModel>& /* model */,
		const Gtk::TreePath& path, bool /* path_currently_selected */)
{
	std::vector<Gtk::TreePath> selected_rows = get_selection()->get_selected_rows();
	if (selected_rows.size() &&
			((selected_rows.front()[0] != path[0] &&
			  selected_rows.front().size() == 2) ||
			 selected_rows.front().size() != path.size()))
		return false;

	return true;
}

// Bring up a popup menu when the playlist treeview is right clicked.
bool PlaylistView::on_button_press_event(GdkEventButton* event)
{
	if (m_Device.IsReady() && event->type == GDK_BUTTON_PRESS)
	{
		Gtk::TreePath path;
		// Popup the playlist right click menu
		if (event->button == 3)
		{
			if (get_path_at_pos(event->x, event->y, path) && !get_selection()->is_selected(path))
			{
				get_selection()->unselect_all();
				get_selection()->select(path);
			}
			Gtk::Menu* playlist_popup =
				(Gtk::Menu*)m_UIManager->get_widget("/PlaylistPopup");
			playlist_popup->popup(event->button, event->time);
			return true;
		}

		// Get ready for the potential to start a drag
		if ((event->button == 1)
				&& !(event->state & (Gdk::SHIFT_MASK | Gdk::CONTROL_MASK))
				&& (event->x > EXPANDER_WIDTH)
				&& get_path_at_pos(event->x, event->y, path))
		{
			if (path.size() == 2)
			{
				m_DragX = event->x;
				m_DragY = event->y;
				if (get_selection()->is_selected(path) &&
						get_selection()->count_selected_rows())
					return true;
			}
		}
		// Don't allow shift selection to leave the current playlist
		else if ((event->button == 1) && (event->state & Gdk::SHIFT_MASK)
				&& get_path_at_pos(event->x, event->y, path))
		{
			std::vector<Gtk::TreePath> selected_rows =
				get_selection()->get_selected_rows();
			if (selected_rows.size() &&
					((selected_rows.front()[0] != path[0] &&
					  selected_rows.front().size() == 2) ||
					 selected_rows.front().size() != path.size()))
				return true;
		}
	}

	return Gtk::TreeView::on_button_press_event(event);
}

// Set the sensitivity of the items in the playlist toolbar appropriately.
void PlaylistView::SelectionChanged()
{
	int num_selected = get_selection()->count_selected_rows();
	int depth = num_selected ?
		std::vector<Gtk::TreePath>(get_selection()->get_selected_rows()).front().size() :
		0;

	m_UIManager->get_action("/PlaylistPopup/RemoveTracks")->
		set_sensitive(num_selected && depth == 2);
	m_UIManager->get_action("/PlaylistPopup/RemoveTracks")->
		set_visible(num_selected && depth == 2);
	m_UIManager->get_action("/PlaylistPopup/RenamePlaylist")->
		set_sensitive(num_selected == 1 && depth == 1);
	m_UIManager->get_action("/PlaylistPopup/RenamePlaylist")->
		set_visible(num_selected == 1 && depth == 1);
	m_UIManager->get_action("/PlaylistPopup/DeletePlaylists")->
		set_sensitive(num_selected && depth == 1);
	m_UIManager->get_action("/PlaylistPopup/DeletePlaylists")->
		set_visible(num_selected && depth == 1);
}

// A row has been grabbed.  Set up the icon for the drag.
void PlaylistView::on_drag_begin(const Glib::RefPtr<Gdk::DragContext>& context)
{
	Glib::RefPtr<Gtk::TreeModel> model;
	std::vector<Gtk::TreePath> selected_cells = get_selection()->get_selected_rows(model);

	int x = 0, y = 0;
	std::vector< Glib::RefPtr<Pango::Layout> > drag_icon_rows;
	for (std::vector<Gtk::TreePath>::iterator selected = selected_cells.begin();
			selected != selected_cells.end(); ++selected)
	{
		Glib::RefPtr<mtp_data_t> track = (*model->get_iter(*selected))[m_StoreColumns.m_Data];
		drag_icon_rows.push_back(create_pango_layout(track->Name()));
		int width, height;
		drag_icon_rows.back()->get_size(width, height);
		x = std::max(x, (width / PANGO_SCALE));
		y = std::max(y, (height / PANGO_SCALE));
	}

	Cairo::RefPtr<Cairo::Surface> surface =
		get_bin_window()->create_similar_surface(Cairo::CONTENT_COLOR,
				x + 30, (y + 6) * selected_cells.size() + 2);

    Cairo::RefPtr<Cairo::Context> cairo_ctx = Cairo::Context::create(surface);

	// Fill the rectangle with a white background
	cairo_ctx->set_source_rgba(1.0, 1.0, 1.0, 1.0); // white
	cairo_ctx->rectangle(1, 1, x + 28, (y + 6) * selected_cells.size());
	cairo_ctx->fill();

	// Draw the text in black
	cairo_ctx->set_source_rgba(0.0, 0.0, 0.0, 1.0); // black
	int count = 0;
	for (std::vector< Glib::RefPtr<Pango::Layout> >::iterator row = drag_icon_rows.begin();
			row != drag_icon_rows.end(); ++row, ++count)
	{
		cairo_ctx->move_to(15, 3 + count * (y + 6));
		(*row)->show_in_cairo_context(cairo_ctx);

		// Draw a gray line between the text lines
		if (count > 0)
		{
			cairo_ctx->save();
			cairo_ctx->set_source_rgba(0.7, 0.7, 0.7, 1.0); // gray
			cairo_ctx->set_line_width(1.0);
			cairo_ctx->move_to(1, count * (y + 6) + 0.5);
			cairo_ctx->line_to(x + 29, count * (y + 6) + 0.5);
			cairo_ctx->stroke();
			cairo_ctx->restore();
		}
	}

	context->set_icon(cairo_ctx->get_target());
}
